/* deriche.c - implementation of deriches recursive smoothing filter */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>


#define		UCHAR		unsigned char

#define		SQR( x )	((x) * (x))
#define		MAX(x,y)	(((x) > (y)) ? (x) : (y))
#define		MIN(x,y)	(((x) < (y)) ? (x) : (y))


#define		NR		512
#define		NC		512

/* function prototypes */
void DericheSmooth( UCHAR *in, UCHAR *out, int nr, int nc, float sigma );

void DericheSmooth( UCHAR *in, UCHAR *out, int nr, int nc, float sigma )
{
float alpha, k;
float a1, a2, a3, a4, a5, a6, a7, a8;
float b1, b2, c1, c2;
float *y1, *y2, *r;
float sum, expalpha;
int m, n, pixel;

	alpha = 5.0 / (2.0 * sqrt( M_PI ) * sigma );
	expalpha = exp( -alpha );
	k = SQR( 1.0 - expalpha ) / ( 1.0 + 2.0 * alpha * expalpha - SQR( expalpha ) );
	a1 = a5 = k;
	a2 = a6 = k * expalpha * (alpha - 1.0);
	a3 = a7 = k * expalpha * (alpha + 1.0);
	a4 = a8 = -k * SQR( expalpha );
	b1 = 2.0 * expalpha;
	b2 = -SQR( expalpha );
	c1 = c2 = 1.0;

	y1 = (float *)malloc( nr * nc * sizeof( float ) );
	y2 = (float *)malloc( nr * nc * sizeof( float ) );
	r  = (float *)malloc( nr * nc * sizeof( float ) );

	/* first pass */
	for( n = 0; n < nc; n++ ) {
		for( m = 0; m < nr; m++ ) {
			sum = a1 * in[m * nc + n];
			if ( n > 0 ) sum += a2 * in[m * nc + (n-1)] + b1 * y1[m * nc + (n-1)];
			if ( n > 1 ) sum += b2 * y1[m * nc + (n-2)];
			y1[m * nc + n] = sum;
		}
	}

	for( n = nc-1; n >= 0; n-- ) {
		for( m = 0; m < nr; m++ ) {
			sum = 0.0;
			if ( n < (nc-1) ) sum += a3 * in[m * nc + (n+1)] + b1 * y2[m * nc + (n+1)];
			if ( n < (nc-2) ) sum += a4 * in[m * nc + (n+2)] + b2 * y2[m * nc + (n+2)];
			y2[m * nc + n] = sum;
		}
	}

	for( n = 0; n < nc; n++ )
		for( m = 0; m < nr; m++ )
			r[m * nc + n] = c1 * (y1[m * nc + n] + y2[m * nc + n]);


	/* second pass */
	for( m = 0; m < nr; m++ ) {
		for( n = 0; n < nc; n++ ) {
			sum = a5 * r[m * nc + n];
			if ( m > 0 ) sum += a6 * r[(m-1) * nc + n] + b1 * y1[(m-1) * nc + n];
			if ( m > 1 ) sum += b2 * y1[(m-2) * nc + n];
			y1[m * nc + n] = sum;
		}
	}

	for( m = nr-1; m >= 0; m-- ) {
		for( n = 0; n < nc; n++ ) {
			sum = 0.0;
			if ( m < (nr-1) ) sum += a7 * r[(m+1) * nc + n] + b1 * y2[(m+1) * nc + n];
			if ( m < (nr-2) ) sum += a8 * r[(m+2) * nc + n] + b2 * y2[(m+2) * nc + n];
			y2[m * nc + n] = sum;
		}
	}

	for( n = 0; n < nc; n++ )
		for( m = 0; m < nr; m++ ) {
			pixel = (int)(c2 * (y1[m * nc + n] + y2[m * nc + n]) + 0.5);
			pixel = MIN( pixel, 255 );
			pixel = MAX( pixel, 0 );
			out[m * nc + n] = pixel;
		}

	free( r );
	free( y2 );
	free( y1 );
}

