/* pnmio.c - read and write ppm/pgm files */

#include <stdio.h>
#include <stdlib.h>
#include "pnmio.h"



#define		YES		-1
#define		NO		0


#define		UCHAR		unsigned char



static int MagicNumber( FILE *file );
static char pbm_getc( FILE *file );
static int pbm_getint( FILE *file );


int pnm_type( char *fname )
{
FILE *fp;
int magic;

	if ( (fp = fopen( fname, "r" )) == NULL ) return YES;
	magic = MagicNumber( fp );
	fclose( fp );
	return magic;
}


int ppm_read( char *fname, UCHAR **r, UCHAR **g, UCHAR **b, int *nr, int *nc )
{
FILE *fp;
int magic;
int max;
UCHAR *rp, *gp, *bp;
int i, n;
int raw;

	if ( (fp = fopen( fname, "r" )) == NULL ) return YES;

	magic = MagicNumber( fp );
	if ( magic != 3 && magic != 6 ) return YES;

	*nc = pbm_getint( fp );
	*nr = pbm_getint( fp );
	if ( *nr < 0 || *nc < 0 ) return YES;
	max = pbm_getint( fp );
	if ( max < 0 ) return YES;

	
	n = *nr * *nc;

	rp = *r = (UCHAR *)malloc( n * sizeof( UCHAR ) );
	gp = *g = (UCHAR *)malloc( n * sizeof( UCHAR ) );
	bp = *b = (UCHAR *)malloc( n * sizeof( UCHAR ) );

	if ( magic == 3 ) {
		for( i = 0; i < n; i++ ) {
			raw = pbm_getint( fp );
			if ( raw < 0 ) return YES;
			*rp++ = raw;

			raw = pbm_getint( fp );
			if ( raw < 0 ) return YES;
			*gp++ = raw;

			raw = pbm_getint( fp );
			if ( raw < 0 ) return YES;
			*bp++ = raw;
		}
	} else {
		for( i = 0; i < n; i++ ) {
			raw = fgetc( fp );
			if ( raw == EOF ) return YES;
			*rp++ = raw;

			raw = fgetc( fp );
			if ( raw == EOF ) return YES;
			*gp++ = raw;

			raw = fgetc( fp );
			if ( raw == EOF ) return YES;
			*bp++ = raw;
		}
	}

	fclose( fp );
	return NO;
}




int ppm_write( char *fname, UCHAR *r, UCHAR *g, UCHAR *b, int nr, int nc )
{
FILE *fp;
int i, n;

	if ( (fp = fopen( fname, "w" )) == NULL ) return YES;

	fprintf( fp, "P6\n%d %d\n255\n", nc, nr );

	n = nr * nc;

	for( i = 0; i < n; i++ ) {
		putc( *r++, fp );
		putc( *g++, fp );
		putc( *b++, fp );
	}

	fclose( fp );
	return NO;
}



int pgm_read( char *fname, UCHAR **im, int *nr, int *nc )
{
FILE *fp;
int magic;
int max;
int i, n;
int raw;
UCHAR *p;

	if ( (fp = fopen( fname, "r" )) == NULL ) return YES;

	magic = MagicNumber( fp );
	if ( magic != 2 && magic != 5 ) return YES;

	*nc = pbm_getint( fp );
	*nr = pbm_getint( fp );
	if ( *nr < 0 || *nc < 0 ) return YES;
	max = pbm_getint( fp );
	if ( max < 0 ) return YES;

	n = *nr * *nc;

	p = *im = (UCHAR *)malloc( n * sizeof( UCHAR ) );

	if ( magic == 2 ) {
		for( i = 0; i < n; i++ ) {
			raw = pbm_getint( fp );
			if ( raw < 0 ) return YES;
			*p++ = raw;
		}
	} else {
		if ( fread( *im, sizeof( UCHAR ), n, fp ) != n ) return YES;
	}

	fclose( fp );
	return NO;
}




int pgm_write( char *fname, UCHAR *im, int nr, int nc )
{
FILE *fp;
int i, n;

	if ( (fp = fopen( fname, "w" )) == NULL ) return YES;

	fprintf( fp, "P5\n" ); 
	fprintf( fp, "%d %d\n255\n", nc, nr );

	n = nr * nc;

	if ( fwrite( im, sizeof( UCHAR ), n, fp ) != n ) return YES;

	fclose( fp );
	return NO;
}




/* ------------------------------------------------------------------
The following code is adapted from netpbm package by Jef Poskanzer.

	Copyright (C) 1988 by Jef Poskanzer.
--------------------------------------------------------------------- */

static int MagicNumber( FILE *file )
{
int ch1, ch2;

	ch1 = fgetc( file );
	if ( ch1 != 'P' ) return 0;
	ch2 = fgetc( file );
	if ( ch2 < '1' || ch2 > '6' ) return 0;
	return ch2 - '0';
}


static char pbm_getc( FILE *file )
{
int ich;
char ch;

	ich = getc( file );
	if ( ich == EOF ) return (char) EOF;
	ch = (char) ich;
    
	if ( ch == '#' ) {
		do {
			ich = getc( file );
			if ( ich == EOF ) return (char) EOF;
			ch = (char) ich;
		} while ( ch != '\n' && ch != '\r' );
	}

	return ch;
}


static int pbm_getint( FILE *file )
{
char ch;
int i;

	do {
		ch = pbm_getc( file );
	} while ( ch == ' ' || ch == '\t' || ch == '\n' || ch == '\r' );

	if ( ch < '0' || ch > '9' ) return -1;

	i = 0;
	do {
		i = i * 10 + ch - '0';
		ch = pbm_getc( file );
        } while ( ch >= '0' && ch <= '9' );

	return i;
}
